/*
 * Copyright (c) 2019 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */
/**
 * @file
 * @brief Model handler for the light switch.
 *
 * Instantiates a Generic OnOff Client model for each button on the devkit, as
 * well as the standard Config and Health Server models. Handles all application
 * behavior related to the models.
 */
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/mesh/proxy.h>
#include <bluetooth/mesh/models.h>
#include <dk_buttons_and_leds.h>
#include "model_handler.h"

/////////////////////////////
// Tranciever Configurations
/////////////////////////////

void recieve_message(struct tranciever_client *chat, uint8_t *msg)
{
	printk("Responds: %s\n", (char *)msg);
}

void send_message(uint16_t addr, uint8_t *msg)
{
	int err = tranciever_message_send(&tranciever, addr, msg);
	if (err)
	{
		// LOG_WRN("Failed to publish message: %d", err);
	}
}

////////////////////////////////////////////

/* Light switch behavior */

/** Context for a single light switch. */
struct button
{
	/** Current light status of the corresponding server. */
	bool status;
	/** Generic OnOff client instance for this switch. */
	struct bt_mesh_onoff_cli client;
};

static void status_handler(struct bt_mesh_onoff_cli *cli,
						   struct bt_mesh_msg_ctx *ctx,
						   const struct bt_mesh_onoff_status *status);

static struct button buttons[] = {
#if DT_NODE_EXISTS(DT_ALIAS(sw0))
	{.client = BT_MESH_ONOFF_CLI_INIT(&status_handler)},
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw1))
	{.client = BT_MESH_ONOFF_CLI_INIT(&status_handler)},
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw2))
	{.client = BT_MESH_ONOFF_CLI_INIT(&status_handler)},
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw3)) && !defined(CONFIG_BT_MESH_LOW_POWER)
	{.client = BT_MESH_ONOFF_CLI_INIT(&status_handler)},
#endif
};

static void status_handler(struct bt_mesh_onoff_cli *cli,
						   struct bt_mesh_msg_ctx *ctx,
						   const struct bt_mesh_onoff_status *status)
{
	struct button *button =
		CONTAINER_OF(cli, struct button, client);
	int index = button - &buttons[0];

	button->status = status->present_on_off;
	dk_set_led(index, status->present_on_off);

	printk("Button %d: Received response: %s\n", index + 1,
		   status->present_on_off ? "on" : "off");
}

static void button_handler_cb(uint32_t pressed, uint32_t changed)
{
	if (!bt_mesh_is_provisioned())
	{
		return;
	}

	if (IS_ENABLED(CONFIG_BT_MESH_LOW_POWER) && (pressed & changed & BIT(3)))
	{
		bt_mesh_proxy_identity_enable();
		return;
	}

	for (int i = 0; i < ARRAY_SIZE(buttons); ++i)
	{
		if (!(pressed & changed & BIT(i)))
		{
			continue;
		}

		struct bt_mesh_onoff_set set = {
			.on_off = !buttons[i].status,
		};
		int err;

		/* As we can't know how many nodes are in a group, it doesn't
		 * make sense to send acknowledged messages to group addresses -
		 * we won't be able to make use of the responses anyway. This also
		 * applies in LPN mode, since we can't expect to receive a response
		 * in appropriate time.
		 */
		if (bt_mesh_model_pub_is_unicast(buttons[i].client.model) &&
			!IS_ENABLED(CONFIG_BT_MESH_LOW_POWER))
		{
			err = bt_mesh_onoff_cli_set(&buttons[i].client, NULL, &set, NULL);
		}
		else
		{
			err = bt_mesh_onoff_cli_set_unack(&buttons[i].client,
											  NULL, &set);
			if (!err)
			{
				/* There'll be no response status for the
				 * unacked message. Set the state immediately.
				 */
				buttons[i].status = set.on_off;
				dk_set_led(i, set.on_off);
			}
		}

		if (err)
		{
			printk("OnOff %d set failed: %d\n", i + 1, err);
		}
	}
}

/* Set up a repeating delayed work to blink the DK's LEDs when attention is
 * requested.
 */
static struct k_work_delayable attention_blink_work;
static bool attention;

static void attention_blink(struct k_work *work)
{
	static int idx;
	const uint8_t pattern[] = {
#if DT_NODE_EXISTS(DT_ALIAS(sw0))
		BIT(0),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw1))
		BIT(1),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw2))
		BIT(2),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw3))
		BIT(3),
#endif
	};

	if (attention)
	{
		dk_set_leds(pattern[idx++ % ARRAY_SIZE(pattern)]);
		k_work_reschedule(&attention_blink_work, K_MSEC(30));
	}
	else
	{
		dk_set_leds(DK_NO_LEDS_MSK);
	}
}

static void attention_on(struct bt_mesh_model *mod)
{
	attention = true;
	k_work_reschedule(&attention_blink_work, K_NO_WAIT);
}

static void attention_off(struct bt_mesh_model *mod)
{
	/* Will stop rescheduling blink timer */
	attention = false;
}

static const struct bt_mesh_health_srv_cb health_srv_cb = {
	.attn_on = attention_on,
	.attn_off = attention_off,
};

static struct bt_mesh_health_srv health_srv = {
	.cb = &health_srv_cb,
};

BT_MESH_HEALTH_PUB_DEFINE(health_pub, 0);

///////////////////////////////////
// Bluetooth structure definition
///////////////////////////////////

#define TRANCIEVER_TEST

// Initialize Tranciever Object
static struct tranciever_handlers tHandlers = {
	.start = NULL,
	.tx_data = send_message,
	.rx_data = recieve_message,
	.reply = NULL // TODO: add something
};

struct tranciever_client tranciever = {
	.handlers = &tHandlers,
};

static struct bt_mesh_model root_model[] = {
	BT_MESH_MODEL_CFG_SRV,
	BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
	BT_MESH_RX_TX_MODEL(&tranciever),
};

static struct bt_mesh_elem elements[] = {
	BT_MESH_ELEM(0, root_model, BT_MESH_MODEL_NONE),
};

/*
static struct bt_mesh_elem elements[] = {

#ifndef TRANCIEVER_TEST

#if DT_NODE_EXISTS(DT_ALIAS(sw0))
	BT_MESH_ELEM(1,
			 BT_MESH_MODEL_LIST(
				 BT_MESH_MODEL_CFG_SRV,
				 BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
				 BT_MESH_MODEL_ONOFF_CLI(&buttons[0].client)),
			 BT_MESH_MODEL_NONE),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw1))
	BT_MESH_ELEM(2,
			 BT_MESH_MODEL_LIST(
				 BT_MESH_MODEL_ONOFF_CLI(&buttons[1].client)),
			 BT_MESH_MODEL_NONE),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw2))
	BT_MESH_ELEM(3,
			 BT_MESH_MODEL_LIST(
				 BT_MESH_MODEL_ONOFF_CLI(&buttons[2].client)),
			 BT_MESH_MODEL_NONE),
#endif
#if DT_NODE_EXISTS(DT_ALIAS(sw3)) && !defined(CONFIG_BT_MESH_LOW_POWER)
	BT_MESH_ELEM(4,
			 BT_MESH_MODEL_LIST(
				 BT_MESH_MODEL_ONOFF_CLI(&buttons[3].client)),
			 BT_MESH_MODEL_NONE),
#endif
	BT_MESH_ELEM(5,
		 BT_MESH_MODEL_LIST(BT_MESH_MODEL_CFG_SRV,
			BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
			BT_MESH_RX_TX_MODEL(&tranciever)),
			BT_MESH_MODEL_NONE),
#endif
// TrancieverTest
#ifdef TRANCIEVER_TEST

	BT_MESH_ELEM(1,
		 BT_MESH_MODEL_LIST(BT_MESH_MODEL_CFG_SRV,
			BT_MESH_MODEL_HEALTH_SRV(&health_srv, &health_pub),
			BT_MESH_RX_TX_MODEL(&tranciever)),
			BT_MESH_MODEL_NONE),

#endif

};
*/

static const struct bt_mesh_comp comp = {
	.cid = CONFIG_BT_COMPANY_ID,
	.elem = elements,
	.elem_count = ARRAY_SIZE(elements),
};

const struct bt_mesh_comp *model_handler_init(void)
{

	static struct button_handler button_handler = {
		.cb = button_handler_cb,
	};

	dk_button_handler_add(&button_handler);
	k_work_init_delayable(&attention_blink_work, attention_blink);

	return &comp;
}
